"""download_uniprot_AS.py: downloads a file from uniprot with the annotation score for proteins of 
a species"""

import os
import urllib.parse
import urllib.request

def download_AS_file(species: str) -> str:
    """downloads the uniprotKB file for a species with the annotation score column

    Arguments:
        species {str} -- the species number (e.g. 9606)
    Returns:
        str -- the filepath where the file was downloaded to
    """
    '''
    url = 'https://www.uniprot.org/uploadlists/'

    params = {
    'from': 'ACC+ID',
    'format': 'tab',
    "organism": species,
    "columns": "id,genes,annotation_score",
    "compress": "no",
    "query": '',
    "contact": "rseymour@wustl.edu"  # NOTE: not sure if this one works
    }
    '''
    url = "https://www.uniprot.org/uniprot/?query=organism:" + species + "&columns=id,genes,annotation_score&format=tab"
    #url = "https://www.uniprot.org/uniprot/?query=organism:" + species + "+AND+columns=id,genes,annotation_score&format=tab&compress=no&email=rseymour@wustl.edu"
    #headers = {"user-agent": "rseymour@wustl.edu"}
    #data = urllib.parse.urlencode(params)
    #data = data.encode('utf-8')
    req = urllib.request.Request(url)
    with urllib.request.urlopen(req) as f:
        response = f.read()
        annot_score_filename = species + "_annot_score.tsv"
        with open(annot_score_filename, 'w') as as_file:
            decoded_response = response.decode("utf-8")
            if decoded_response[:5] == "Entry":
                as_file.write(decoded_response)
            else:
                return "ERROR: Invalid identifier"
        #print(response.decode('utf-8'))
    return os.path.join(os.getcwd(), annot_score_filename)

if __name__ == "__main__":
    print(download_AS_file("8296"))
#EOF


#!/usr/bin/env python

"""parse_generic.py: parses and rolls up the results from any search engine"""

from collections import defaultdict
import csv
import os.path
from re import findall, finditer, match
import sys
from time import sleep

from .proteoSushi_constants import cleave_rules
from .ps_utilities import load_pepdict


def get_PTMs(sky_filename: str) -> list:
    """Gets the list of PTMs available through parsing the generic output

    Arguments:
        sky_filename {str} -- the name of the generic output file
    Returns:
        list -- a list of PTMs in the output file
    """
    with open(sky_filename, 'r') as sky_output:
        tsv_reader = csv.reader(sky_output, quotechar='"')
        header = next(tsv_reader)
        try:
            seq_index = header.index("Peptide Sequence")
        except ValueError:
            return -4
        mod_index = -1
        try:
            mod_index = header.index("Peptide Modified Sequence") #NOTE: apparently, this changes, so check here.
        except ValueError:
            try:
                mod_index = header.index("Modified Sequence")
            except ValueError:
                print("\033[91m {}\033[00m".format("No peptide modified sequence column detected in the Skyline output file."))
                print("\033[91m {}\033[00m".format("Please add or modify header with the name \"Peptide Modified Sequence\"\n"))
                return -4
        
        mod_list = []
        for row in tsv_reader:
            # This grabs the PTMs in each sequence and builds a list of all of them
            sequence = row[seq_index].replace("L","I")
            mod_seq = row[mod_index]
            mods = findall(r"(\w?\[.+?\])|(\w?\(.+?\(?.\)?\))", mod_seq)

            #If there are no PTMs in the sequence
            if len(mods) < 1:
                return -3  # This will be there error listed below, but handled in the GUI
                raise ValueError("A sequence in the Peptide Modified Sequence column is missing PTMs")

            for mod in mods[0]:
                if mod == '':
                    continue
                stripped_mod = mod
                if not stripped_mod in mod_list:
                    mod_list.append(stripped_mod)
    return mod_list

def create_mod_dict(sky_filename: str, user_PTMs: list) -> dict:
    """Parses the generic output to retrieve the modification info

    Arguments:
        sky_filename {str} -- the name of the generic output file
        user_PTMs {list} -- a list of PTMs that the user wants to analyze
    Returns:
        dict -- a dictionary of sequences with the type and location of modifications
    """
    mod_dict = {}
    with open(sky_filename, 'r') as sky_output:
        tsv_reader = csv.reader(sky_output, quotechar='"')
        header = next(tsv_reader)
        seq_index = header.index("Peptide Sequence")
        mod_index = -1
        try:
            mod_index = header.index("Peptide Modified Sequence") #NOTE: apparently, this changes, so check here.
        except ValueError:
            try:
                mod_index = header.index("Modified Sequence")
            except ValueError:
                print("\033[91m {}\033[00m".format("No peptide modified sequence column detected in the Skyline output file."))
                print("\033[91m {}\033[00m".format("Please add or modify header with the name \"Peptide Modified Sequence\"\n"))
                sys.exit()
        
        for row in tsv_reader:
            # This grabs the PTMs in each sequence and builds a list of all of them
            sequence = row[seq_index].replace("L","I")
            #if sequence == "FACAVVCIQK":
            #    print("here")
            mod_seq = row[mod_index]
            mods = findall(r"(\w?\[.+?\])|(\w?\(.+?\(?.\)?\))", mod_seq)
            if len(mods) < 1:
                raise ValueError("A sequence in the Peptide Modified Sequence column is missing PTMs")

            breaks = finditer(r"(\w?\[.+?\])|(\w?\(.+?\(?.\)?\))", mod_seq)
            cut_sites = []
            for breakp in breaks:
                cut_sites.append(breakp.start())
            # This fixes the indices of each mod to connect with the unmodified sequence
            correction = 0
            fixed_indices = []
            num_of_mods = 0
            for site in cut_sites:
                fixed_indices.append(site - correction)  # TODO: Check this!
                correction += len(mods[num_of_mods][0]) - 1
                num_of_mods += 1
            
            i = 0
            while i < len(cut_sites):
                if not mods[i][0] in user_PTMs:
                    i += 1
                    continue
                try:
                    mod_dict[mod_seq].append(tuple((mods[i][0], fixed_indices[i])))  # TODO: Check This!
                except KeyError:
                    mod_dict[mod_seq] = [tuple((mods[i][0], fixed_indices[i]))]
                i += 1
    return mod_dict

def __chooseHit(genes_positions: list, mito_genes: list, annotDict: dict) -> list:
    """chooses which of the matched sequences to use. If there is one mito gene, it will be that one.
    If there are more than one non-mito, annotation score decides. If there are more than one mito, annotation score decides.
    Arguments:
        genes_positions {list} -- a list of tuples with gene and position info
        mito_genes {list} -- a list of strings of mito gene names
        annotDict {dict} -- a dictionary connecting genes to annotation scores
    Returns:
        bool -- whether the match(es) is(are) target genes
        *and*
        str -- the gene name of the chosen match
        int -- the start position of the chosen match
        *or*
        list -- a list of tuples with the gene name and start position of each match
    """
    mitoTups = list()
    nonMitoTups = list()
    for tup in genes_positions:
        if tup[0].upper() + '\n' in mito_genes:
            #print("Mito gene prioritized!")
            mitoTups.append(tup)
        else:
            nonMitoTups.append(tup)
    if len(mitoTups) == 1:
        return True, mitoTups[0]
    elif len(mitoTups) > 1:
        return True, __chooseTup(mitoTups, annotDict)
    elif len(nonMitoTups) == 1:
        return False, nonMitoTups[0]
    elif len(nonMitoTups) > 1:  # Only non-mito proteins (and more than 1)
        return False, __chooseTup(nonMitoTups, annotDict)
    assert False, "ERROR: chooseHit had 0 tuples sent in!"
    #return False, None, None

def __chooseTup(tuples: list, annotDict: dict) -> list:
    """chooses which tuple to return from a list
    Arguments:
        tuples {list} -- a list of tuples with gene and start position info
        annotDict {dict} -- a dictionary connecting genes to annotation scores
    Returns:
        str -- the gene name of the chosen match
        int -- the start position of the chosen match
    """
    highestScore = 0
    high2Score = 0
    highest = None
    #print(f"third {annotDict["MUG2"]}")
    for tup in tuples:
        if not tup[2] in annotDict:
            continue
        if int(annotDict[tup[2]]) > highestScore:  # If the current match is has the highest score, set it
            highestScore = int(annotDict[tup[2]])
            highest = tup
        elif int(annotDict[tup[2]]) > high2Score:  # If the current match is 2nd highest score, set it
            high2Score = int(annotDict[tup[2]])
    if highestScore > high2Score:
        return highest
    else:  # If there are tied high scores for annotation
        sharedPeps = list()
        sharedPeps.append(highest)
        for tup in tuples:  # Cycles through the matches and chooses the ones with the highest annotation score
            if not tup[2] in annotDict:
                continue
            elif int(annotDict[tup[2]]) == highestScore and tup[2] != highest[2]:
                sharedPeps.append(tup)
        if len(sharedPeps) == 1:
            return sharedPeps[0]
        return sharedPeps
        ### Here is the paralog code - Not running currently ###
        '''with open("Paralogs_rat_UniprotIDs_genes.csv", 'r') as par:
            paralogDict = dict()
            par.readline()
            for line in par:  # Make the paralog dictionary so we can check whether ties are paralogs
                if line.split(',')[0] in paralogDict:
                    paralogDict[line.split(',')[0]].append(line.strip().split(',')[1])
                else:
                    paralogDict[line.split(',')[0]] = [line.strip().split(',')[1]]
                if line.split(',')[1] in paralogDict:
                    paralogDict[line.split(',')[1]].append(line.strip().split(',')[0])
                else:
                    paralogDict[line.split(',')[1]] = [line.strip().split(',')[0]]
            paralogs = list()
            paralogs.append(highest)
            for tup in tuples:  # Cycles through the matches and chooses the ones with the highest annotation score if they are paralogs
                if not tup[0] in annotDict:
                    continue
                elif int(annotDict[tup[0]]) == highestScore and tup[0] != highest[0]:
                    if tup[0] in paralogDict and highest[0] in paralogDict[tup[0]]:
                        paralogs.append(tup)
            if len(paralogs) > 1:
                return paralogs
            else:
                return None, None'''
'''
def __prompt_PTMs(PTMs: list) -> list:
    """prompts the user for which PTMs should be used
    Arguments:
        PTMs {list} -- a list of PTM names
    Returns:
        list -- a list of PTMs specified by the user
    """
    print("\033[96m {}\033[00m".format("Possible PTMs for analysis include:"))
    print("\033[93m {}\033[00m".format("\n ".join([f"[{i}] {ptm}" for i, ptm in enumerate(PTMs)])))
    print("\033[96m {}\033[00m".format("Enter the number for each of the PTMs of interest separated by a comma."))
    ptm_indices = input("\033[96m {}\033[00m".format("For example: 0,1,2\n"))
    if match(r"{^0-9|,}", ptm_indices):  # r"[0-9](?:,[0-9])*"
        print("\033[91m {}\033[00m".format("Invalid input, try again!\n"))
        sleep(.5)
        return __prompt_PTMs(PTMs)  # NOTE: the end of this recursion is getting it right
    try:
        mod_PTMs = [PTMs[int(x.strip())] for x in ptm_indices.strip().split(',')]
    except (ValueError, IndexError):
        print("\033[91m {}\033[00m".format("Invalid input, try again!\n"))
        sleep(.5)
        return __prompt_PTMs(PTMs)
    return mod_PTMs
'''

def compile_data_generic(search_engine_filepath: str, user_PTMs: list) -> list:
    """Takes the lists and dictionaries needed to parse files

    Arguments:
        search_engine_filepath {str} -- the filepath for the generic search engine output
        user_PTMs {list} -- the list of PTMs that the user will use for the analysis
    Returns:
        int -- index of unmodified sequence
        int -- index of modified sequence
        dict -- pep_dict
        dict -- mod_dict
        int -- intensity_values
        file -- mito_genes (file)
        str -- sky_filename
        dict -- mascot var mod dict (blank)
    """
    sky_filename = search_engine_filepath #__prompt_file()
    mod_dict = create_mod_dict(sky_filename, user_PTMs)
    PTMs = user_PTMs#__prompt_PTMs(mod_list)
    #Collecting the PTM analysis list from the user
    # NOTE: is it possible to not use PTMs for this analysis?
    #cleave_rule, maxMissed = __promptCleavenMissed()
    #pep_dict = load_pepdict(cleave_rule, maxMissed)  # TODO: make it find the max missed from file
    #annotDict = __load_annotDict("uniprot_rat_annotScore_ids.csv")
    #annotDict = __load_annotDict("uniprot-proteome_human_annotScore.tsv")
    #gene_results = []
    #dup_check_set = set()
    #threshold = 0.01 #What should this be?
    skyFile = open(sky_filename, 'r')
    tsv_reader = csv.reader(skyFile, quotechar='"')
    header = next(tsv_reader)
    sequence = header.index("Peptide Sequence")
    try:
        pms = header.index("Peptide Modified Sequence") #NOTE: apparently, this changes, so check here.
    except ValueError:
        try:
            pms = header.index("Modified Sequence")
        except ValueError:
            print("\033[91m {}\033[00m".format("No peptide modified sequence column detected in the Skyline output file."))
            print("\033[91m {}\033[00m".format("Please add or modify header with the name \"Peptide Modified Sequence\"\n"))
            sys.exit()
    #UNPID = -1
    #try:
    #    UNPID = header.index("Protein Accession")
    #except ValueError:
    #    print("\033[91m {}\033[00m".format("No uniprot ID column detected in the Skyline output file."))
    #    print("Proceeding regardless.\n")
    intensity_values = [i for i, h in enumerate(header) if "intensit" in h.lower()]  # NOTE: This makes the assumption that the intensity columns are directly after the charge column
    if len(intensity_values) == 0:
        intensity_values = -1
    #psm_contributions = defaultdict(int)
    #unmatchedPeps = 0
    #missingPTM = 0
    #totalSeqs = 0
    #unmatchedSequences = []
    #intensity_dict = dict()
    return sequence, pms, mod_dict, intensity_values, sky_filename, None
#EOF

#/home/rob/Documents/Held_Lab/191010_Rollup/SkylineOutput_GloCys055_EXP0107u_RED.csv


"""parse_mascot.py: functions used to parse the mascot file and read it into memory"""

import csv
from collections import defaultdict
import os
from re import match
from time import sleep

from .ps_utilities import parse_mascot, load_pepdict #most of the parsing actually happens here 

'''
def __promptFile() -> str:
    """Prompts the user for the file name
    Returns:
        str -- the filename
    """
    print("\033[96m {}\033[00m".format("Please provide the filename of the Mascot file"))
    print("\033[96m {}\033[00m".format("For example: \nC:/experiment1/Combined/F011000.csv, or"))
    filename = input("\033[96m {}\033[00m".format("/home/[user]/Documents/experiment1/F011000.csv\n"))
    if not os.path.exists(filename):
        print("\033[91m {}\033[00m".format("Invalid path, try again!\n"))
        sleep(.5)
        return __promptFile()
    return filename
'''

def __create_mod_dict(filename: str, mod_ids: list, var_mod_map: dict) -> dict:
    """Makes the PTM dictionary that connects a peptide sequence with its modifications
    Arguments:
        filename {str} -- the name of the output file
        mod_ids {list} -- a list of mod ids that mascot uses and the user chose
        var_mod_map {dict} -- connects the PTM to the Mascot index

    Returns:
        dict -- a dictionary of sequences with the type and location of modifications
    """
    modDict = {}
    inv_mod_map = {v:k for k, v in var_mod_map.items()}
    with open(filename, 'r') as mascot_output:
        file_reader = csv.reader(mascot_output, quotechar='"')
        header = next(file_reader)
        sequence = header.index("pep_seq")
        #pep_mod_seq_index = header.index("")
        var_mods = header.index("pep_var_mod_pos")

        for row in file_reader:
            #Mascot has duplicate rows; check for these
                
            pep_seq = row[sequence].replace("L","I")
                
            mods = row[var_mods]

            if not mods:
                continue
            new_seq = pep_seq
            pep_mod_seq = mods.split('.')[1]
            i = len(pep_mod_seq) - 1
            inv_mod_dict = {v:k for k, v in var_mod_map.items()}
            while i >= 0:
                if pep_mod_seq[i] != "0":  # If there is a mod
                    new_seq = new_seq[:i+1] + '(' + inv_mod_dict[pep_mod_seq[i]] + ')' + new_seq[i+1:]
                i -= 1
            pep_mod_seq = new_seq

            if mods:
                n, center, c = [x for x in mods.split(".")]
                for i, aa in enumerate(list(center)): #split mods; note this excludes termini
                    if aa in mod_ids:
                        try:
                            if not tuple((str(aa), i)) in modDict[pep_mod_seq]:
                                modDict[pep_mod_seq].append(tuple((inv_mod_map[aa], i)))
                        except KeyError:
                            modDict[pep_mod_seq] = [tuple((inv_mod_map[aa], i))]
    return modDict

'''
def __promptPTMs(PTMs: list) -> list:
    """prompts the user for which PTMs should be used
    Arguments:
        PTMs {list} -- a list of PTM names
    Returns:
        list -- a list of PTMs specified by the user
    """
    print("\033[96m {}\033[00m".format("Possible PTMs for analysis include:"))
    print("\033[93m {}\033[00m".format("\n ".join([f"[{i}] {ptm}" for i, ptm in enumerate(PTMs)])))
    print("\033[96m {}\033[00m".format("Enter the number for each of the PTMs of interest separated by a comma."))
    ptmIndices = input("\033[96m {}\033[00m".format("For example: 0,1,2\n"))
    if match(r"{^0-9|,}", ptmIndices):  # r"[0-9](?:,[0-9])*"
        print("\033[91m {}\033[00m".format("Invalid input, try again!\n"))
        sleep(.5)
        return __promptPTMs(PTMs)  # NOTE: the end of this recursion is getting it right
    try:
        modPTMs = [PTMs[int(x.strip())] for x in ptmIndices.strip().split(',')]
    except (ValueError, IndexError):
        print("\033[91m {}\033[00m".format("Invalid input, try again!\n"))
        sleep(.5)
        return __promptPTMs(PTMs)
    return modPTMs
'''

def compile_data_mascot(search_engine_filepath: str, PTMs: list) -> list:
    """Takes the lists and dictionaries needed to parse files

    Arguments:
        search_engine_filepath {str} -- the filepath to the mascot output file
        PTMs {list} -- the list of PTMs that the user has chosen to use
    Returns:
        int -- index of unmodified sequence
        int -- index of modified sequence
        dict -- mod_dict
        int -- intensity_start 
        str -- data_filename
        dict -- dictionary of the mods to the index
    """
    ###Set input file; perhaps change to an argument and put in function

    #in_file = "MascotExportForRollup-UniHumanRefFASTA-20190731_4Rob.csv"
    in_file = search_engine_filepath #__promptFile()
    enzyme, quant_range, var_mod_map, missed_cleaves = parse_mascot(in_file)

    ###Change this to a variable for later, perhaps let the user choose from var_mod_map
    ###See code below
    #mod_for_quant = "IYn_CleavedPC_TMT6 (C)".lower()
    '''
    if not mod_for_quant: #make this a choice for later
        print(*var_mod_map.items(), sep = '\n')
        mod_choice = input("Choose which modification you want to use for quant (use number): ")
        mod_for_quant = [key for key, value in var_mod_map.items() if value == mod_choice][0]
        print(mod_for_quant)
        input()
    '''
    mods_for_quant = PTMs #__promptPTMs(list(var_mod_map.keys()))

    #assert mod_for_quant in var_mod_map, "Make sure mod is in original file"
    #mod_for_quant_id = var_mod_map.get(mod_for_quant)

    #pep_dict = load_pepdict(enzyme, missed_cleaves)

    #lociDict = dict()

    #print("Quant range", quant_range)#Just see if it found the correct quant columns

    #gene_results = dict()
    #dup_check_set = set()
    #threshold = 0.01 #Make this a mod, too

    input_filename = "quant_parsed.csv"
    input_file =  open(input_filename, 'r')
    file_reader = csv.reader(input_file)
    header = next(file_reader)

    sequence = header.index('pep_seq') #assert this exists later on
    var_mods = header.index('pep_var_mod_pos')
    #score = header.index('pep_expect')
    #print(sequence, var_mods)
    #failedSeqs = 0
    #totalSeqs = 0

    #psm_contributions = defaultdict(int) #count number of psms that end up being rolled up

    mod_dict = __create_mod_dict(input_filename, [var_mod_map[mod] for mod in mods_for_quant], var_mod_map)

    if quant_range[1] <= quant_range[0]:
        return sequence, var_mods, mod_dict, None, input_filename, var_mod_map

    return sequence, var_mods, mod_dict, quant_range[0], input_filename, var_mod_map
#EOF


"""run_proteoSushi.py: The starting point to run proteoSushi"""

from .proteosushi_gui import run_gui
from .combine_intensities import rollup
'''
def run_proteoSushi(search_engine: str, search_engine_filepath: str, use_target_list: bool, 
                    target_list_filepath: str, max_missed_cleavages: int, protease: str, fdr_threshold: float):
    """starts proteoSushi rollup when called by the GUI

    Arguments:
        search_engine {str} -- "maxquant", "mascot", or "generic"
        search_engine_filepath {str} -- the filepath for search engine output
        use_target_list {bool} -- whether to use the target list to prioritize matches
        target_list_filepath {str} -- the filepath for the target gene list
        max_missed_cleavages {int} -- the maximum allowed missed cleavages in a peptide
        protease {str} -- the protease used previously to cleave the proteins in the sample
        fdr_threshold {float} -- threshold used for the pep_expect and PEP columns in mascot/maxquant
    """
    pass
'''
if __name__ == "__main__":
    run_gui()
    '''
    rollup(
        search_engine="generic",
        search_engine_filepath="/home/rob/Documents/Held_Lab/ProteoSushi/proteosushi/examples/EGFR_Skyline_data.csv",
        use_target_list=False,
        target_list_filepath="",
        max_missed_cleavages=3,
        protease="trypsin/p",
        fdr_threshold=.01,
        use_quant=False,
        user_PTMs=["C[+57]"],
        proteome_fasta_filepath="/home/rob/Documents/Held_Lab/ProteoSushi/proteosushi/examples/fastas/Uni-Hum-Ref-20141022.fasta",
        intensity_method="",
        add_annotation=True,
        species_id="9606"
    )
    '''
#EOF


#!/usr/bin/env python

"""parse_MaxQuant.py: using helper functions, the script parses MaxQuant result files"""

from collections import defaultdict
import csv
import os.path
from re import match
from time import sleep

from .ps_utilities import load_pepdict, parse_maxquant_summary

def parse_evidence(filename: str) -> dict:
    """parses the evidence file and builds a dictionary for PTMs
    Arguments:
        filename {str} -- name of the evidence file with the location
    Returns:
        dict -- a dictionary of the PTMs with sequence as key and a list of tuples with 
        PTM/location as value
        list -- a list of the PTMs we can use for the analysis
    """
    mod_dict = dict()
    PTMs = list()
    with open(filename, 'r') as evidence_file:
        tsv_reader = csv.reader(evidence_file, delimiter='\t', quotechar='"')
        header = next(tsv_reader)
        seq_index = header.index("Sequence")
        #modIndices = [i for i, s in enumerate(header) if "Probabilities" in s]
        modified_index = header.index("Modified sequence")
        mods_index = header.index("Modifications")
        #PTMs = [ptm.split(" Probabilities")[0] for ptm in header if "Probabilities" in ptm]
        start_index = header.index([ptm for ptm in header if "Score Diffs" in ptm][-1]) + 1
        end_index = header.index("Missed cleavages")
        PTMs = header[start_index:end_index]
        for row in tsv_reader:
            pep_mods = dict()
            sequence = row[seq_index].replace('L', 'I')
            #if sequence == "MAPACQIIR":
            #    print("peptide found")
            for PTM in PTMs:
                modified_sequence = row[modified_index].strip('_')
                mods_B4_mod = 0
                # Grabs the first two letters of PTM that maxquant uses in mod_seq
                abr_ptm = PTM[:2].lower()  # NOTE: I could possibly imagine this not working in certain circumstances...
                # while this PTM is in the sequence we are looking at and listed among the PTMs for this peptide
                while PTM in row[mods_index] and abr_ptm in modified_sequence:  # This seems a little sketchy. What if there are two different PTMs with the same first 2 letters?
                    other_PTMs = PTMs.copy()
                    other_PTMs.remove(PTM)
                    # Gets the index of the PTM within the sequence (that is right here XX(|ptm)XX)
                    for other_ptm in other_PTMs:
                        mod_index = modified_sequence.index(abr_ptm)
                        abr_other_ptm = other_ptm[:2].lower()
                        while (abr_other_ptm in modified_sequence and 
                               modified_sequence.index(abr_other_ptm) < mod_index):
                            #This takes the ptms, one by one, until each before is removed
                            mods_B4_mod += 1
                            ptm_index = modified_sequence.index(abr_other_ptm)
                            modified_sequence = (modified_sequence[:ptm_index]
                                               + modified_sequence[ptm_index + len(abr_other_ptm):])  # NOTE: removes the mod, but not the parenthesis
                    mod_index = modified_sequence.index(abr_ptm)
                    pep_mods[modified_sequence.index(abr_ptm) - (mods_B4_mod * 2)] = PTM
                    modified_sequence = (modified_sequence[:mod_index]
                                       + modified_sequence[mod_index + len(abr_ptm):])
                    mods_B4_mod += 1
            for loc in sorted(pep_mods):
                try:
                    if not tuple((pep_mods[loc], loc - 2)) in mod_dict[row[modified_index].strip('_')]:
                        mod_dict[row[modified_index].strip('_')].append(tuple((pep_mods[loc],
                                                        loc - 2)))
                except KeyError:
                    mod_dict[row[modified_index].strip('_')] = [tuple((pep_mods[loc],
                                                loc - 2))]
    return mod_dict, PTMs
'''
def __prompt_dir() -> str:
    """Prompts the user for the directory with the MaxQuant files
    Returns:
        str -- the path to the directory
    """
    print("\033[96m {}\033[00m".format("Please provide the path to the MaxQuant files."))
    print("\033[96m {}\033[00m".format("For example: C:/experiment1/Combined/combined/txt/, or"))
    MQ_dir = input("\033[96m {}\033[00m".format("/home/[user]/Documents/experiment1/txt/\n"))
    if not os.path.exists(f"{MQ_dir}evidence.txt"):
        print("\033[91m {}\033[00m".format("Invalid path, try again!\n"))
        sleep(.5)
        return __prompt_dir()
    return MQ_dir

def __prompt_PTMs(PTMs: list) -> list:
    """prompts the user for which PTMs should be used
    Arguments:
        PTMs {list} -- a list of PTM names
    Returns:
        list -- a list of PTMs specified by the user
    """
    print("\033[96m {}\033[00m".format("Possible PTMs for analysis include:"))
    print("\033[93m {}\033[00m".format("\n ".join([f"[{i}] {ptm}" for i, ptm in enumerate(PTMs)])))
    print("\033[96m {}\033[00m".format("Enter the number for each of the PTMs of interest separated by a comma."))
    ptm_indices = input("\033[96m {}\033[00m".format("For example: 0,1,2\n"))
    if match(r"{^0-9|,}", ptm_indices):  # r"[0-9](?:,[0-9])*"
        print("\033[91m {}\033[00m".format("Invalid input, try again!\n"))
        sleep(.5)
        return __prompt_PTMs(PTMs)  # NOTE: the end of this recursion is getting it right
    try:
        mod_PTMs = [PTMs[int(x.strip())] for x in ptm_indices.strip().split(',')]
    except (ValueError, IndexError):
        print("\033[91m {}\033[00m".format("Invalid input, try again!\n"))
        sleep(.5)
        return __prompt_PTMs(PTMs)
    return mod_PTMs
'''
def compile_data_maxquant(search_engine_filepath: str, user_PTMs: list) -> list:
    """Takes the lists and dictionaries needed to parse files
    Returns:
        int -- index of unmodified sequence
        int -- index of modified sequence
        dict -- mod_dict
        int -- intensity_start (maybe just skyline)
        str -- evidence_filename
        list -- PTMs the user has chosen to be used
    """
    MQ_dir = search_engine_filepath #__prompt_dir()
    sum_file = os.path.join(MQ_dir, "summary.txt")
    enzyme, max_missed_cleaves = parse_maxquant_summary(sum_file)
    evidence_filename = os.path.join(MQ_dir, "evidence.txt")
    mod_dict, PTMs = parse_evidence(evidence_filename)  # This grabs the PTMs from the evidence file
    PTMs = user_PTMs#__prompt_PTMs(PTMs)
    evidence_file = open(evidence_filename, 'r')
    tsv_reader = csv.reader(evidence_file, delimiter='\t', quotechar='"')
    header = next(tsv_reader)
    sequence = header.index("Sequence")
    mod_seq = header.index("Modified sequence")
    reporter_intensities = [i for i, h in enumerate(header) if "intensit" in h.lower() and 
                                                               not "max intensity m/z" in h.lower()]
    #reporter_intensity = header.index("Intensity")
    evidence_file.close()
    return sequence, mod_seq, mod_dict, reporter_intensities, evidence_filename, PTMs
#EOF


"""sparql.py: takes the results from proteoSushi and annotates it using the sparql API to access uniprot"""

__author__ = "Rob Seymour, Arshag Mooradian"
__email__ = "rseymour@wustl.edu"

from io import StringIO
import json

import csv
import pandas as pd
from re import findall
import requests
import sys  # Remove later
from time import sleep
import urllib


# Global constants
ENDPOINT = "https://sparql.uniprot.org"

def sparql_json_to_df(sparql_json_dict: dict):
    """Converts the messy json format into a dataframe

    Arguments:
        sparql_json_dict {dict} -- a json object retrieved through requests.get query
    Returns:
        sparql_df {pd.DataFrame} -- dataframe containing the results
    """
    
    cols = sparql_json_dict["head"]["vars"]

    temp_row_list = [] # list holding rows
    for row in sparql_json_dict["results"]["bindings"]:
        item = []
        for c in cols:
            item.append(row.get(c, {}).get("value"))
        temp_row_list.append(item)

    sparql_df = pd.DataFrame(temp_row_list, columns = cols)
    return sparql_df


def sparql_request(unpid_site_list: list):
    """Sends the request to uniprot to get data on each of the sites
    Arguments:
        unpidSiteList {list} -- a list of tuples with uniprot ID and site of PTM
    Returns:
    """
    unpid_site_list_str = ""
    if not unpid_site_list:
        return ""
    for tup in unpid_site_list:
        if len(tup) >= 2 and tup[1] == '0':
            tup[1] = '1'
        #if tup[0] == "P60709" or tup[0] == "P63261":
        #    print("Multiple")
        if len(tup[0]) >= 5:
            unpid_site_list_str += f"(uniprotkb:{tup[0]} {tup[1]})\n"
    # TODO: Delete this later
    #with open("sparqllist.tsv", 'w') as sparqllist:
    #    sparqllist.write(unpid_site_list_str)
    #print(unpid_site_list_str)


    prefix="""PREFIX uniprotkb: <http://purl.uniprot.org/uniprot/>
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX faldo: <http://biohackathon.org/resource/faldo#>
PREFIX up: <http://purl.uniprot.org/core/>
PREFIX ec: <http://purl.uniprot.org/enzyme/>"""


    new_entry_length_pos_query = prefix + """
SELECT
    ?entry
    ?position ### New the position of the C that was given in the values
    (STRLEN(?iupac) AS ?lengthOfSequence)
    ?begin
    ?end
    (SUBSTR(?iupac, ?begin, ?end - ?begin + 1) AS ?regionOfInterest)
FROM <http://sparql.uniprot.org/uniprot>
WHERE {
    VALUES (?entry ?position) {""" + unpid_site_list_str + """}
    ?entry up:sequence ?sequence .
    ?annotation up:range ?range ;
               a ?type .
    ?range faldo:begin
        [ faldo:position ?begin ; faldo:reference ?sequence ] ;
            faldo:end
        [ faldo:position ?end ; faldo:reference ?sequence ] .
    FILTER (?begin <= ?position && ?position <= ?end)
    # get the IUPAC AAs associated with the identifier
    ?sequence rdf:value ?iupac .
    # get the AA subsequence of the annotation as a new variable
} ORDER BY DESC(?entry) ASC(?position)"""

    query_stripped_entry_pos = prefix + """
SELECT
    ?entry
    ?position ### New the position of the C that was given in the values
    ?lengthOfSequence
    ?begin
    ?end
    ?regionOfInterest ### subsequence of the annotation that had the C
FROM <http://sparql.uniprot.org/uniprot>
WHERE {
     VALUES (?entry ?position) {""" + unpid_site_list_str + """}
     ?entry up:sequence ?sequence .
     ?annotation up:range ?range1, ?range2;
                a ?type .
     ?range1 faldo:begin
         [ faldo:position ?begin ; faldo:reference ?sequence ] .
     ?range2 faldo:end
         [ faldo:position ?end ; faldo:reference ?sequence ] .
     FILTER (?begin <= ?position && ?position <= ?end)
     # get the IUPAC AAs associated with the identifier
     ?sequence rdf:value ?iupac .
} ORDER BY DESC(?entry) ASC(?position)"""

    query_entry_length_position = prefix + """
SELECT
    ?entry
    ?position ### New the position of the C that was given in the values
    ?lengthOfSequence
    ?begin
    ?end
    ?regionOfInterest ### subsequence of the annotation that had the C
FROM <http://sparql.uniprot.org/uniprot>
WHERE {
    VALUES (?entry ?position) {""" + unpid_site_list_str + """
    }
    ?entry up:sequence ?sequence .
    ?annotation up:range ?range ;
               a ?type .
    ?range faldo:begin
        [ faldo:position ?begin ; faldo:reference ?sequence ] ;
            faldo:end
        [ faldo:position ?end ; faldo:reference ?sequence ] .
    FILTER (?begin <= ?position && ?position <= ?end)
    # get the IUPAC AAs associated with the identifier
    ?sequence rdf:value ?iupac .
    # get the AA subsequence of the annotation as a new variable
    BIND(SUBSTR(?iupac, ?begin, ?end - ?begin + 1) AS ?regionOfInterest)
    # get length
    ?sequence rdf:value ?iupac .
    BIND(strlen(?iupac) AS ?lengthOfSequence)
} ORDER BY DESC(?entry) ASC(?position)"""


    query_entry_subcellular = prefix + """
SELECT
    ?entry 
    ?location
FROM <http://sparql.uniprot.org/uniprot>
WHERE {
    VALUES (?entry ?position) {""" + unpid_site_list_str + """
    }
    ?entry up:annotation ?subAnnotation .
    ?subAnnotation a up:Subcellular_Location_Annotation .
    ?subAnnotation up:locatedIn/up:cellularComponent ?location .
} ORDER BY DESC(?entry)"""


    query_catalytic = prefix + """
SELECT
    ?entry
    ?position
    ?catalyicActivity
FROM <http://sparql.uniprot.org/uniprot>
WHERE {
    VALUES (?entry ?position) {""" + unpid_site_list_str + """
    }
   ?entry up:sequence ?sequence .
    ?entry up:annotation ?caAnnotation .
    ?caAnnotation a up:Catalytic_Activity_Annotation .
    ?caAnnotation up:catalyticActivity ?ca .
} ORDER BY DESC(?entry) ASC(?position)"""


    query_ec_rhea_type = prefix + """
SELECT
    ?entry
    ?position
    ?ec
    ?rhea
    ?type
    ?comment
FROM <http://sparql.uniprot.org/uniprot>
WHERE {
    VALUES (?entry ?position) {""" + unpid_site_list_str + """
    }
    ?entry up:sequence ?sequence .
    ?entry up:annotation ?annotation .
    ?annotation up:range ?range ;
               a ?type .
    ?range faldo:begin
        [ faldo:position ?begin ; faldo:reference ?sequence ] ;
            faldo:end
        [ faldo:position ?end ; faldo:reference ?sequence ] .
    FILTER (?begin <= ?position && ?position <= ?end)
    ?entry up:annotation ?caAnnotation .
    ?caAnnotation a up:Catalytic_Activity_Annotation .
    ?caAnnotation up:catalyticActivity ?ca .

    OPTIONAL {
        ?ca up:enzymeClass ?ec .
    }

    OPTIONAL {
    ?ca up:catalyzedReaction ?rhea .
    }

    OPTIONAL {
        ?annotation rdfs:comment ?comment .
    }

#Order desc by entry and ascending by position afterwards
} ORDER BY DESC(?entry) ASC(?position)"""

    # Grabs the annotation by parts to maximize the amount we receive from the uniprot server
    region_annot = request_annot(new_entry_length_pos_query)
    #print(region_annot.head(15))
    #if region_annot.empty:
    region_annot_stripped = request_annot(query_stripped_entry_pos)
    #print(region_annot_stripped.head(15))
    if len(region_annot.index) < len(region_annot_stripped.index):  # This should check to see if there are rows missing in the full query
        region_annot = region_annot.append(region_annot_stripped)
        #print(region_annot.head(30))
        region_annot.drop_duplicates(subset=["entry", " position", " begin", " end"], keep="first", inplace=True)
    #print(region_annot.head(15))
    #catalytic_annot = request_annot(query_catalytic)
    subcell_annot = request_annot(query_entry_subcellular)
    #print(subcell_annot.head(15))
    extras_annot = request_annot(query_ec_rhea_type)
    # Creates blank dataframes if uniprot did not return that info
    if region_annot is None:
        region_annot = pd.DataFrame(columns=["entry", " position", " lengthOfSequence", " begin", " end", " regionOfInterest"])
    else:
        region_annot.dropna(how="all", inplace=True)
    #if catalytic_annot is None:
    #    catalytic_annot = pd.DataFrame(columns=["entry", " position", " catalyticActivity"])
    #else:
    #    catalytic_annot.dropna(how="any", inplace=True)
    if subcell_annot is None:
        subcell_annot = pd.DataFrame(columns=["entry", " location"])
    else:
        subcell_annot.dropna(how="any", inplace=True)
    if extras_annot is None:
       extras_annot = pd.DataFrame(columns=["entry", " position", " ec", " rhea", " type", " comment"])
    else:
        extras_annot.dropna(how="all", inplace=True)  # TODO: I will likely need to change this back to all later

    # Full outer joins the annotations to preserve all info possible
    full_annot = region_annot
    del(region_annot)
    try:
        #full_annot = full_annot.merge(catalytic_annot, how="outer", on=["entry", " position"])
        #del(catalytic_annot)
        full_annot = full_annot.merge(subcell_annot, how="outer", on="entry")
        #print(full_annot.head(15))
        del(subcell_annot)
        full_annot = full_annot.merge(extras_annot, how="outer", on=["entry", " position"])
        del(extras_annot)
    except KeyError:
        print(full_annot)
        #print(isinstance(full_annot, pd.DataFrame))
        if isinstance(full_annot, pd.DataFrame) and "502 Proxy Error" in full_annot.iloc[1][0]:
            return 502
        sys.exit()
    return full_annot

def request_annot(query: str, attempts_left=10):
    """Sends the query to uniprot and checks the output

    Arguments:
        query {str} -- the query with the sites to annotate
        attempts_left {int} -- the number of requests that can be sent out
    Returns:
        pandas.dataframe -- the returned annotation
    """
    headers = {"user-agent": "rseymour@wustl.edu"}
    time_to_sleep = 5
    try:
        r = requests.post(ENDPOINT, data = {"format": "csv", "query": query}, headers = headers)
        if "<!DOCTYPE html SYSTEM \"about:legacy-compat\">" in r.text:  # This just means it returned a 500 error
            if attempts_left > 0:
                sleep(time_to_sleep)
                print("\033[93m {}\033[00m".format(f"Attempts used: {11 - attempts_left}"))
                return request_annot(query, attempts_left - 1)
            else:
                return None  # TODO: remove this once Uniprot fixes their stuff
                raise pd.errors.ParserError("Ran out of Uniprot accession attempts")
        csv_file = StringIO(r.text)
        sparql_from_csv_df = pd.read_csv(csv_file)
    except (pd.errors.ParserError, requests.exceptions.ConnectionError):
        if attempts_left > 0:
            sleep(time_to_sleep)
            print("\033[93m {}\033[00m".format(f"Attempts used: {11 - attempts_left}"))
            return request_annot(query, attempts_left - 1)
        else:
            return None  # TODO: remove this once Uniprot fixes their stuff
            raise pd.errors.ParserError("Ran out of Uniprot accession attempts")
    return sparql_from_csv_df

def process_sparql_output(output_df, sparql_dict: dict) -> list:
    """Process the output from uniprot to make it consistent with the rollup output

    Arguments:
        output_df {pandas.dataframe} -- the string output from uniprot, may have multiple
        sparql_dict {dict} -- the growing dictionary of uniprot annotations
    Returns:
        list -- a list of annotations with the data properly separated
    """

    def __parse_sparql_line(line: str) -> str:
        """changes the complicated uniprot annotation output into a cleaner line

        Arguments:
            line {str} -- a line from the output
        Returns:
            str -- a cleaned up version of that same line
        """
        #entry,lengthOfSequence,catalyicActivity,location,ec,rhea,type,comment,position,begin,end,regionOfInterest
        new_str = list()
        for chunk in line[1]:
            if isinstance(chunk, str) and "http://purl.uniprot.org/uniprot/" in chunk:
                new_str.append(chunk.replace("http://purl.uniprot.org/uniprot/", ""))
            elif isinstance(chunk, str) and "http://purl.uniprot.org/core/" in chunk:
                new_str.append(chunk.replace("http://purl.uniprot.org/core/", ""))
            elif isinstance(chunk, str) and "http://purl.uniprot.org/locations/" in chunk:
                new_chunk = chunk.replace("http://purl.uniprot.org/locations/", "")
                new_str.append(subcellular_location_dict[new_chunk.zfill(4)])  # Converts the number location to descriptor
            elif isinstance(chunk, str) and "http://purl.uniprot.org/enzyme/" in chunk:
                new_chunk = chunk.replace("http://purl.uniprot.org/enzyme/", "")
                new_str.append(ec_id_to_description_dict[new_chunk])
            elif isinstance(chunk, str) and "^^<http://www.w3.org/2001/XMLSchema#int>" in chunk:
                new_str.append(findall(r"(\d+?)\^", chunk)[0])
            else:
                new_str.append(str(chunk))
        return '\t'.join(new_str)

    output_list = list()
    comments_dict = dict()
    #sparql_dict = dict()  # This gets used by the main program to connect these annotations to the rest of the data.
    output_lines = output_df
    if len(output_lines) == 1 or (not isinstance(output_lines, str) and output_lines.empty):
        print("Failed to get annotations")
        return output_list, sparql_dict
    
    try:
        position_index = output_lines.columns.get_loc(" position")
    except (AttributeError, KeyError):
        print("Missing the Position column")
        return output_list, sparql_dict

    # get the subcellular location from UniProt
    def retrieve_uniprot_subcellular_location():
        '''Retrieve subcellular location information from UniProt directly'''
        r = requests.get('https://www.uniprot.org/locations/?format=tab')
        subcellular_location_df = pd.read_csv(StringIO(r.text), sep = '\t')  # TODO: potential ParserError
        subcellular_location_df["location_id_reformatted"] = (
                                                             subcellular_location_df["Subcellular location ID"]
                                                             .str.split("-")
                                                             .str[1]
                                                             )
        return subcellular_location_df
    
    subcellular_location_df = retrieve_uniprot_subcellular_location()
    subcellular_location_dict = dict(subcellular_location_df[["location_id_reformatted", "Alias"]].values.tolist())

    # Fetches a file from expasy to deal with the 'ec' column
    try:
        r = urllib.request.urlopen("ftp://ftp.expasy.org/databases/enzyme/enzyme.dat", "enzyme.dat")
    except urllib.error.URLError:
        sleep(10)
        r = urllib.request.urlopen("ftp://ftp.expasy.org/databases/enzyme/enzyme.dat", "enzyme.dat")
    enzyme_dat_list = r.read().decode("utf-8").split('\n')

    # Turn enzyme.dat into a dictionary for mapping
    ec_id_to_description_dict = dict()
    for line_number, line in enumerate(enzyme_dat_list):
        if line.startswith("ID "):
            current_id = line[5:]
            current_description = enzyme_dat_list[line_number + 1][5:]
            ec_id_to_description_dict[current_id] = current_description

    #entry,position,lengthOfSequence,begin,end,regionOfInterest,location,ec,rhea,type,comment
    #header = output_lines.colnames()
    #print(output_lines.columns)
    
    #i = 1
    repeat_index = 3
    #while i < len(output_lines):
    for output_line in output_lines.iterrows():  # TODO: optimize by vectorizing
        output_line = __parse_sparql_line(output_line)
        # This should only have 1 iteration and is used to properly split the tsv type line
        for output_split in csv.reader([output_line], delimiter='\t', quotechar='"'):
            key = ','.join([*output_split[:repeat_index]])
            try:
                comments_dict[key] += output_split[repeat_index:]
            except KeyError:
                comments_dict[key] = output_split[repeat_index:]
        #i += 1
    
    for key in comments_dict:
        output_list.append(key + ',' + ','.join(comments_dict[key]))
        sparql_dict[ key.split(',')[0] + '|' + key.split(',')[1] ] = key.split(',') + comments_dict[key] #unpid, pos
    #print(f"sparql_dict has {sparql_dict}")
    return output_list, sparql_dict
#EOF


"""ps_utilities.py: helper functions used by parsers in proteoSushi"""

__author__ = "Arshag Mooradian, Rob Seymour"
__email__ = "rseymour@wustl.edu"

import pickle
import csv
from collections import defaultdict
import ntpath
import os

from .ps_digest import digest, fasta_producer, cleave_rule_determination


def parse_maxquant_summary(infile: str) -> list:
    """Pulls out information needed from the MaxQuant files
    Arguments:
        infile {str} -- the name of the MaxQuant summary file that will be used
    Returns:
        str -- the name of the enzyme used (usually trypsin)
        int -- the maximum allowed number of missed cleavages
    """
    mqfile = open(infile, 'r')
    #outfile = open("mq_parsed.csv", 'w', newline='')
    #assert infile.endswith(".txt")
    tsvReader = csv.reader(mqfile, delimiter='\t', quotechar='"')
    #csvWriter = csv.writer(outfile)
    enzyme = ""
    PTMs = []
    maxMissedCleaves = 0
    # From proteinGroups: Need Intensity[DM], PTM data[GS-GX], Potential Contaminant(?)[GK]
    # From summary: Enzyme[C], Variable Modifications[H][I-J?], Max. missed cleavages[O]
    # From parameters: Modifications included in protein quantification,
    header = next(tsvReader)  # Skips the header line
    row = next(tsvReader)
    enzyme = row[header.index("Enzyme")]
    assert enzyme != ""
    #PTMs = row[header.index("Variable modifications")].split(';')  # I might just do this in the evidence file
    maxMissedCleaves = int(row[header.index("Max. missed cleavages")])
    mqfile.close()
    return enzyme.lower(), maxMissedCleaves

def parse_mascot(in_file: str) -> list:
    """Pulls out enzyme, quantitative range, and variable modifications, along
    with the quant data from a Mascot file.
    Arguments:
        in_file {str} -- a string for the name of the mascot file
    Returns:
        list -- enzyme {str}, quant_range {tuple}, var_mod_map {dict},
        missed_cleaves {int}
    """
# We need to parse Mascot file first, because raw Mascot output is difficult to work with
    print(f"Parsing file {in_file}.")
    with open(in_file) as f, \
        open('quant_parsed.csv', 'w', newline='') as out_file:
        # Quantitative section of Mascot file
        assert in_file.endswith('.csv'), "File must be a csv"
        file_reader = csv.reader(f)
        out_writer = csv.writer(out_file)
        enzyme = ""
        missed_cleaves = -1
        quant_range = None
        var_mod_map = dict()
        for row in file_reader:
            if row:  # lots of blank rows
                # Pull data we need
                if row[0] == "Enzyme":
                    enzyme = row[1].lower()
                elif row[0] == "Variable modifications" and '--' not in row[1]:
                    var_mods = row[1].split(',')
                    var_mod_map = {val.lower():
                                   str(pos+1) for pos, val in enumerate(var_mods)}
                elif row[0] == "Maximum Missed Cleavages":
                    missed_cleaves = int(row[1])
                elif row[0] == "prot_hit_num" or row[0] == "prot_acc":
                    # This should be last elif; start of quantitative info
                    header = row
                    header = [i for i in header if i]
                    second_row = next(file_reader)  # need to normalize headers
                    #print(header)
                    #print(second_row)
                    quant_range = (len(header), len(second_row)-1)  #range where there will be quant; appears to skip some
                    header.extend(['SampleQuant' 
                                   for i in range(len(second_row) - len(header))])
                    out_writer.writerows([header, second_row])
                    for row in file_reader:
                        out_writer.writerow(row)
                    break  # Exit loop as we have all of the header information
        print("Finished Parsing")
        if quant_range is None:
            return -5, None, None, None
        return enzyme, quant_range, var_mod_map, missed_cleaves


def load_pepdict(proteome_fasta_filepath: str, enzyme: str, missed_cleaves: int) -> dict:
    """Looks for pep_dict; if it doesn't find, it looks for a FASTA to make one.
    Fasta would be in current directory but this can be changed.

    Arguments:
        proteome_fasta_filepath {str} -- the filepath for the proteome fasta file
        enzyme {str} -- protease used in the sample digestion
        missed_cleaves {int} -- the max number of allowed missed cleavages
    Returns:
        dict -- a dictionary that 
    """
    pep_dict = None
    #wd = os.getcwd()
    fasta_dir = os.path.dirname(os.path.abspath(proteome_fasta_filepath))
    pep_dict_file = f"{proteome_fasta_filepath.split('.')[0]}.pepdict"
    #print(os.listdir(fasta_dir))
    if not any([x == ntpath.basename(pep_dict_file) for x in os.listdir(fasta_dir)]):
        print("Pepdict not found in the FASTA directory. Trying to generate from the FASTA file")
        digest_if_needed(proteome_fasta_filepath, enzyme, missed_cleaves)
    '''
    pep_dicts = 0
    for f in os.listdir(wd):
        if f.endswith('.pepdict'):
            pep_dicts += 1
            pep_dict_file = f
    assert pep_dicts == 1, \
        f"Found {pep_dicts} pepdict files. Need exactly 1 in {wd}."
        '''
    
    print(f"Loading {pep_dict_file} into memory")
    with open(pep_dict_file, 'rb') as f:
        pep_dict = pickle.load(f)
    print("\033[95m {}\033[00m".format("Peptide Dictionary Loaded."))
    return pep_dict


def digest_if_needed(proteome_fasta_filepath: str, enzyme: str, mcleave: bool, minlen=5, maxlen=55, 
                     m_excise="Both"):
    """Runs digest and dumps pickle file (dict)"""
    seq_list = []  # (gene, organism, sequence, unpid)
    seq_list = fasta_producer(proteome_fasta_filepath, seq_list)
    rule = cleave_rule_determination(enzyme)
    pep_dict = defaultdict(set)  # NOTE: this won't work for large fasta/sequence databases as it's in ram; ~1GB for human
    for i in seq_list:
        gene, _organism, seq, unpid, protein_name = i
        if m_excise == "Both":
            # Run digest results twice and combine; inefficient but gets job done
            digest_results_true = digest(seq, minlen, maxlen, mcleave, True,
                                         li_swap=True, rule_to_use=rule,
                                         reverse=False)
            digest_results_false = digest(seq, minlen, maxlen, mcleave, False,
                                          li_swap=True, rule_to_use=rule,
                                          reverse=False)
            digest_results = set().union(digest_results_true,
                                         digest_results_false)
        else:
            digest_results = digest(seq, minlen, maxlen, mcleave, m_excise,
                                    li_swap=True, rule_to_use=rule,
                                    reverse=False)
        for result in digest_results:
            peptide, position = result
            peptide = peptide.replace("L", "I")
            pep_dict[peptide].add((gene, position, unpid, protein_name))  # careful with site specificity here
    with open(f"{proteome_fasta_filepath.split('.')[0]}.pepdict", "wb") as w1:
        pickle.dump(pep_dict, w1)
    del pep_dict  # Free memory for later



"""combine_output_annotations.py: since annotation is still imperfect, I combine
the patchy output into a file that has as much as possible
"""

import os
import pandas as pd

main_rollup_file = None
annotations_dict = dict()
for f in os.listdir(os.getcwd()):
    if "generic_rollup" in f:
        print(f"Combining {f}")
        rollup_output = pd.read_csv(open(f, 'rU'), engine="python")
        if main_rollup_file is None:
            main_rollup_file = rollup_output
            '''
            for line in main_rollup_file.iterrows():
                try:
                    if line["entry"]:
                        annotations_dict[f"{line["Site"]}|{line["Peptide Modified Sequence"]}"] = line["entry",]
                except KeyError:
                    continue
                '''
        else:
            main_rollup_file.update(rollup_output, overwrite=False)
            '''
            for line in rollup_output.iterrows():
                try:
                    annotations_dict[f"{line["Site"]}|{line["Peptide Modified Sequence"]}"]
                except KeyError:
                    if line["entry"]:
                        annotations_dict[f"{line["Site"]}|{line["Peptide Modified Sequence"]}"] = line["entry",]
                    continue
                '''

main_rollup_file.to_csv("generic_combined_rollup.csv")
#EOF


#!/usr/bin/env python

"""combine_intensities.py: takes data from the parser and combines the mods and intensities"""

import csv
import os
import pandas as pd
from re import finditer, match, findall
from time import sleep

from .download_uniprot_AS import download_AS_file
from .parse_mascot import compile_data_mascot
from .parse_MaxQuant import compile_data_maxquant, parse_evidence
from .parse_generic import compile_data_generic, get_PTMs
from .sparql import process_sparql_output, sparql_request
from .proteoSushi_constants import cleave_rules, annotation_type_dict, secondary_annotations
#from ruputilities import load_pepdict, parse_mascot, parse_maxquant_summary
from .ps_utilities import parse_mascot, load_pepdict, parse_maxquant_summary

'''
def __promptCleavenMissed() -> list:
    """Prompts the user for the cleavage rule and max missed cleavages
    Returns:
        str -- the filename
        int -- number of allowed misses
    """
    cleave_rules = {
        # C terminal proteases
        'trypsin/p': (r'[RK]', 'c'),
        'trypsin!p': (r'[RK](?!P)', 'c'),
        'lys-c': (r'[K]', 'c'),
        # N terminal proteases
        'asp-n': (r'[D]', 'n'),
        'asp-nc': (r'[DC]', 'n'),
        'lys-n': (r'[K]', 'n')
    }
    print("\033[96m {}\033[00m".format("Please provide the cleavage rule."))
    cleaveRule = input("\033[96m {}\033[00m".format("Examples include 'trypsin/p', 'trypsin!p', 'lys-c', 'asp-n', asp-nc', 'lys-n'\n"))
    if not cleaveRule in cleave_rules:
        print("\033[91m {}\033[00m".format("Not an accepted format; please try again.\n"))
        sleep(.5)
        return __promptCleavenMissed()
    missed_cleavages = __promptMissed()
    return cleaveRule, missed_cleavages

def __promptMissed() -> int:
    missed_cleavages = None
    try:
        missed_cleavages = int(input("\033[96m {}\033[00m".format("Please provide the number of allowed missed cleavages.\n")))
    except Exception:
        print("\033[91m {}\033[00m".format("Not an accepted format; please try again.\n"))
        sleep(.5)
        return __promptMissed()
    if missed_cleavages < 0:
        print("\033[91m {}\033[00m".format("Max missed cleavages must be a positive number; please try again.\n"))
        sleep(.5)
        return __promptMissed()
    return missed_cleavages
'''
def __chooseHit(genes_positions: list, target_genes: list, annot_dict: dict, use_target: bool) -> list:
    """Chooses which of the matched sequences to use. If there is one target gene, it will be that one.
    If there are more than one non-target, annotation score decides. If there are more than one target, annotation score decides.
    
    Arguments:
        genes_positions {list} -- a list of tuples with gene and position info
        target_genes {list} -- a list of strings of target gene names
        annotDict {dict} -- a dictionary connecting genes to annotation scores
    Returns:
        bool -- whether the match(es) is(are) target genes
        *and*
        str -- the gene name of the chosen match
        int -- the start position of the chosen match
        *or*
        list -- a list of tuples with the gene name and start position of each match
    """
    target_tups = list()
    nontarget_tups = list()
    for tup in genes_positions:
        if use_target and tup[0].upper() + '\n' in target_genes:
            #print("Mito gene prioritized!")
            target_tups.append(tup)
        else:
            nontarget_tups.append(tup)
    if len(target_tups) == 1:
        return True, target_tups[0]
    elif len(target_tups) > 1:
        return True, __chooseTup(target_tups, annot_dict)
    elif len(nontarget_tups) == 1:
        return False, nontarget_tups[0]
    elif len(nontarget_tups) > 1:  # Only non-target proteins (and more than 1)
        return False, __chooseTup(nontarget_tups, annot_dict)
    assert False, "ERROR: chooseHit had 0 tuples sent in!"
    #return False, None, None

def __chooseTup(tuples: list, annot_dict: dict) -> list:
    """chooses which tuple to return from a list
    Arguments:
        tuples {list} -- a list of tuples with gene and start position info
        annotDict {dict} -- a dictionary connecting genes to annotation scores
    Returns:
        str -- the gene name of the chosen match
        int -- the start position of the chosen match
    """
    highestScore = 0
    high2Score = 0
    highest = None
    #print(f"third {annotDict["MUG2"]}")
    for tup in tuples:
        if not tup[2] in annot_dict:
            continue
        if int(annot_dict[tup[2]]) > highestScore:  # If the current match is has the highest score, set it
            highestScore = int(annot_dict[tup[2]])  # NOTE: This is not working in maxquant, it never gets here
            highest = tup
        elif int(annot_dict[tup[2]]) > high2Score:  # If the current match is 2nd highest score, set it
            high2Score = int(annot_dict[tup[2]])
    if highestScore > high2Score:
        return highest
    else:  # If there are tied high scores for annotation
        no_scores = True
        sharedPeps = list()
        sharedPeps.append(highest)
        for tup in tuples:  # Cycles through the matches and chooses the ones with the highest annotation score
            if not tup[2] in annot_dict:
                continue
            elif int(annot_dict[tup[2]]) == highestScore and tup[2] != highest[2]:
                no_scores = False
                sharedPeps.append(tup)
        if not sharedPeps[0] is None and len(sharedPeps) == 1:  # If there is a top scorer, send that one
            return sharedPeps[0]
        elif no_scores:  # If none of the proteins have a score, send all of them
            return tuples
        return sharedPeps
        ### Here is the paralog code - Not running currently ###
        '''with open("Paralogs_rat_UniprotIDs_genes.csv", 'r') as par:
            paralogDict = dict()
            par.readline()
            for line in par:  # Make the paralog dictionary so we can check whether ties are paralogs
                if line.split(',')[0] in paralogDict:
                    paralogDict[line.split(',')[0]].append(line.strip().split(',')[1])
                else:
                    paralogDict[line.split(',')[0]] = [line.strip().split(',')[1]]
                if line.split(',')[1] in paralogDict:
                    paralogDict[line.split(',')[1]].append(line.strip().split(',')[0])
                else:
                    paralogDict[line.split(',')[1]] = [line.strip().split(',')[0]]
            paralogs = list()
            paralogs.append(highest)
            for tup in tuples:  # Cycles through the matches and chooses the ones with the highest annotation score if they are paralogs
                if not tup[0] in annotDict:
                    continue
                elif int(annotDict[tup[0]]) == highestScore and tup[0] != highest[0]:
                    if tup[0] in paralogDict and highest[0] in paralogDict[tup[0]]:
                        paralogs.append(tup)
            if len(paralogs) > 1:
                return paralogs
            else:
                return None, None'''


def __clean_pep_seq(rule: tuple, pep_mod_seq: str, user_PTMs: list) -> str:
    """Removes the missed cleavages without selected PTMs

    Arguments:
        rule {tuple} -- contains info about the protease
        pep_mod_seq {str} -- the peptide sequence with modifications
        user_PTMs {list} -- a list of PTMs that the user has selected for Analysis
    Returns:
        str -- the cleaned peptide
    """
    breaks = finditer(rule[0], pep_mod_seq)
    cut_sites = []
    cut_peptides = []
    for breakp in breaks:
        cut_sites.append(breakp.start())

    last_site = 0
    if rule[1].lower() == 'c':
        for i in cut_sites:
            cut_peptides.append(pep_mod_seq[last_site:i+1])
            last_site = i + 1
        cut_peptides.append(pep_mod_seq[last_site:])
    elif rule[1].lower() == 'n':
        for i in cut_sites:
            cut_peptides.append(pep_mod_seq[last_site:i])
            last_site = i
        cut_peptides.append(pep_mod_seq[last_site:])
    # This gets the range of the first and last pep_slice with 
    start = 0
    while start < len(cut_peptides):
        if any(user_PTM in cut_peptides[start] for user_PTM in user_PTMs):
            break
        else:
            start += 1
    
    end = len(cut_peptides) - 1
    while end > start:
        if any(user_PTM in cut_peptides[end] for user_PTM in user_PTMs):
            break
        else:
            end -= 1

    if start == end:
        new_pep_mod_seq = cut_peptides[0]
    else:
        new_pep_mod_seq = ''.join(cut_peptides[start:end])
    return new_pep_mod_seq

def consolidate_sequence(new_pep_mod_seq: str, new_user_PTMs: list) -> str:
    """Changes the modified peptide sequence to only have the relevant PTMs for indexing

    Arguments:
        new_pep_mod_seq {str} -- the modified peptide sequence
        new_user_PTMs {list} -- the PTMs selected by the user
    Returns:
        str -- the modified peptide sequence with irrelevant PTMs removed (for indexing)
    """
    try:
        mods = findall(r"(\[.+?\])|(\(.+?\(?.\)?\))", new_pep_mod_seq).remove('')
    except ValueError:
        mods = findall(r"(\[.+?\])|(\(.+?\(?.\)?\))", new_pep_mod_seq)
    # If there is a PTM in the sequence that isn't in the user list
    #if all([mod[0] in new_user_PTMs for mod in mods]):
    #    return new_pep_mod_seq

    for mod in mods:
        if not mod[1][1:-1] in new_user_PTMs:
            new_pep_mod_seq = new_pep_mod_seq.replace(mod[1], "")
    return new_pep_mod_seq

def __add_intensity(intensity_dict: dict, new_pep_mod_seq: str, genes: list, site: int, 
                    intensities: list) -> list:
    """Adds the intensity of the site to the dictionary, if entry exists, adds to the numbers

    Arguments:
        intensity_dict {dict} -- a dictionary connecting the pep/gene/site to intensity
        new_pep_mod_seq {str} -- the consolidated, modified sequence of the peptide
        genes {list} -- a list of gene names associated with this peptide
        site {int} -- the position of the mod in the protein
        intensities {list} -- 
    Returns:
        dict -- the intensity_dict following the update
        bool -- True if a new entry was added, False if not
    """
    to_adds = []
    for gene in genes:
        key = f"{new_pep_mod_seq}|{gene.upper()}|{str(site)}"
        #if gene.upper() == "SQRDL":
        #    print("SQRDL_379")
        if key in intensity_dict:
            old_intensities = intensity_dict[key]
            new_intensities = list()
            new_ints = list()
            i = 0

            # Combines the new values for the peak
            while i < len(intensities):
                old_int = old_intensities[i]
                if not old_int or "#N/A" == old_int or "NaN" == old_int or not old_int:
                    old_int = 0

                new_int = intensities[i]
                if not new_int or "#N/A" == new_int or "NaN" == new_int or not new_int:
                    new_int = 0

                new_ints.append(new_int)
                new_intensities.append(float(old_int) + float(new_int))
                i += 1

            # Adds to the number of combined peptides for averaging later (if needed)
            if any(new_ints):
                new_intensities.append(int(old_intensities[i]) + 1)  # NOTE: what does it mean if this is out of range?
            else:
                new_intensities.append(int(old_intensities[i]))

            # Replaces the combined numbers in the dictionary
            intensity_dict[key] = new_intensities
            to_adds.append(False)

        else:
            new_intensities = []
            for new_int in intensities:
                if "#N/A" == new_int:
                    new_int = 0
                new_intensities.append(new_int)

            # Adds in the number of combined peptide peaks (1 so far)
            new_intensities.append(1)
            # Adds a new entry for these peaks
            intensity_dict[key] = new_intensities
            to_adds.append(True)

    return intensity_dict, any(to_adds)

def __load_annot_dict(annot_file: str) -> dict:
    """loads the annotation score file to make a dict that connects gene to annotation score

    Arguments:
        annot_file {str} -- the filename for the annotation score file
    Returns:
        dict -- a dictionary connecting genes to annotation scores
    """
    annot_dict = {}
    if annot_file == "":
        return annot_dict
    with open(annot_file, 'r') as r1:
        tsv_reader = csv.reader(r1, delimiter='\t')
        header = next(tsv_reader)
        uniprot_ID = header.index("Entry")
        annot_score = header.index("Annotation")
        for row in tsv_reader:
            annot_dict[row[uniprot_ID]] = row[annot_score][0]
    return annot_dict


def parse_output(search_engine: str, search_engine_filepath: str) -> list:
    """Parses the search engine output and grabs info to return to the gui

    Arguments:
        search_engine {str} -- "maxquant", "mascot", or "generic"
        search_engine_filepath {str} -- the filepath for search engine output
    Returns:
        int -- max missed cleavages allowed
        str -- protease used to cleave the proteins
    """
    rollup_file = search_engine
    if rollup_file == "generic":
        missed_cleavages = -1
        enzyme = ""
        PTMs = get_PTMs(search_engine_filepath)
        if PTMs == -3:
            return -3, None, None  # "A sequence in the Peptide Modified Sequence column is missing PTMs"
        if PTMs == -4:
            return -4, None, None
    elif rollup_file == "maxquant":
        MQ_dir = search_engine_filepath
        sum_file = os.path.join(MQ_dir, "summary.txt")
        enzyme, missed_cleavages = parse_maxquant_summary(sum_file)
        mod_dict, PTMs = parse_evidence(os.path.join(MQ_dir, "evidence.txt"))  # This grabs the PTMs from the evidence file
    elif rollup_file == "mascot":
        enzyme, quant_range, var_mod_map, missed_cleavages = parse_mascot(search_engine_filepath)
        if enzyme == -5:
            return -5, None, None
        PTMs = list(var_mod_map.keys())
    else:  # It shouldn't be able to get here
        assert False, "Not a valid file from search"

    return missed_cleavages, enzyme, PTMs


def __compress_annotations(annotation_list: list) -> list:
    """Takes a list of annotations and makes a single combined group

    Arguments:
        annotation_list {list} -- a list of annotations comprising 1 or more groups
    Returns:
        list -- the compressed list of annotations
    """
    #entry,position,lengthOfSequence,begin,end,regionOfInterest,location,ec,rhea,type,comment(,begin,end,...)
    # NOTE: you may need to update these numbers if you add or delete a column
    begin_index = 3
    end_index = 4
    type_index = 9
    comment_index = 10
    length_uniprot_annotations = 8  # The number of 
    # Grabs the entry, position, and lengthofsequence columns
    new_annotations = annotation_list[:begin_index]
    # Compresses the begin and end columns into a range column
    new_annotations.append(f"{annotation_list[begin_index]}-{annotation_list[end_index]}")
    # Adds the rest of the columns (up to type)
    new_annotations += annotation_list[end_index+1:type_index]
    # Adds a blank spot for each annotation (as listed in proteosushi_constants.py)
    new_annotations += [""]*len(annotation_type_dict)

    try:  # Attempts to put the comment in the appropriate column using the type as reference
        new_annotations[annotation_type_dict[annotation_list[type_index]] + length_uniprot_annotations] = annotation_list[comment_index]
    except KeyError:
        if annotation_list[type_index] == "nan":
            pass
        else:
            new_annotations[annotation_type_dict["Other"] + length_uniprot_annotations] = annotation_list[comment_index]

    # NOTE: you will need to update these numbers if you add or remove a column from query
    new_begin_index = begin_index - 3
    new_end_index = end_index - 3
    range_index = 3
    region_index = 4
    location_index = 5
    ec_index = 6
    rhea_index = 7
    new_type_index = 8
    new_comment_index = 9
    secondary_structure = ""
    #index_fix = 1  # Fixes the index after I combine the begin and end into range
    i = 1  # Basically which group it is on
    while (i * length_uniprot_annotations) + 3 < len(annotation_list):  # Essentially, we are moving through the annotation list and compressing it to new annotations
        if not f"{annotation_list[new_begin_index+(i*length_uniprot_annotations) + 3]}-{annotation_list[new_end_index+(i*length_uniprot_annotations)+3]}" in new_annotations[range_index].split(','):
            new_annotations[range_index] += f",{annotation_list[new_begin_index+(i*length_uniprot_annotations) + 3]}-{annotation_list[new_end_index+(i*length_uniprot_annotations)+3]}"
        if not f"{annotation_list[region_index-2+(i*length_uniprot_annotations) + 3]}" in new_annotations[region_index].split(','):
            new_annotations[region_index] += f",{annotation_list[region_index-2+(i*length_uniprot_annotations) + 3]}"
        #if annotation_list[catalytic_index-2+(i*length_uniprot_annotations)+3] != new_annotations[catalytic_index]:
        #    if annotation_list[catalytic_index-2+(i*length_uniprot_annotations)+3] == "nan":
        #        new_annotations[catalytic_index] += ','
        #    else:
        #        new_annotations[catalytic_index] += f",{annotation_list[catalytic_index-2+(i*length_uniprot_annotations)+3]}"
        if not f"{annotation_list[location_index-2+(i*length_uniprot_annotations)+3]}" in new_annotations[location_index].split(','):
            new_annotations[location_index] += f",{annotation_list[location_index-2+(i*length_uniprot_annotations)+3]}"
        if not f"{annotation_list[ec_index-2+(i*length_uniprot_annotations)+3]}" in new_annotations[ec_index].split(','):
            new_annotations[ec_index] += f",{annotation_list[ec_index-2+(i*length_uniprot_annotations)+3]}"
        if not f"{annotation_list[rhea_index-2+(i*length_uniprot_annotations)+3]}" in new_annotations[rhea_index].split(','):
            new_annotations[rhea_index] += f",{annotation_list[rhea_index-2+(i*length_uniprot_annotations)+3]}"

        try:  # Attempts to put the comment in the appropriate column using the type as reference
            current_comment = annotation_list[new_comment_index-2+(i*length_uniprot_annotations)+3]
            current_type = annotation_list[new_type_index-2+(i*length_uniprot_annotations)+3]

            if (not current_comment in new_annotations[annotation_type_dict[current_type] + length_uniprot_annotations] and
                not current_type in new_annotations[annotation_type_dict[current_type] + length_uniprot_annotations]):
                if new_annotations[annotation_type_dict[current_type]
                                + length_uniprot_annotations]:
                    if current_comment != "nan":
                        new_annotations[annotation_type_dict[current_type]
                                        + length_uniprot_annotations] += f",{current_comment}"
                else:
                    if current_comment == "nan":
                        new_annotations[annotation_type_dict[current_type]
                                        + length_uniprot_annotations] += f"{current_type.replace('_Annotation', '')}"
                    else:
                        new_annotations[annotation_type_dict[current_type]
                                        + length_uniprot_annotations] += f"{current_comment}"
        except KeyError:
            if current_type == "nan":
                pass
            else:
                # If the current comment is not in the Other cell
                if current_type in secondary_annotations:
                    secondary_structure = current_type.replace("_Annotation", "")
                elif (not current_comment in new_annotations[annotation_type_dict["Other"] + length_uniprot_annotations] and
                    not current_type in new_annotations[annotation_type_dict["Other"] + length_uniprot_annotations]):
                    if new_annotations[annotation_type_dict["Other"] + length_uniprot_annotations]:
                        if current_comment != "nan":
                            new_annotations[annotation_type_dict["Other"] 
                                            + length_uniprot_annotations] = f",{current_type.replace('_Annotation', '')}: {current_comment}"
                    else:
                        if current_comment == "nan":
                            new_annotations[annotation_type_dict["Other"] 
                                            + length_uniprot_annotations] = f"{current_type.replace('_Annotation', '')}"
                        else:
                            new_annotations[annotation_type_dict["Other"] 
                                            + length_uniprot_annotations] = f"{current_type.replace('_Annotation', '')}: {current_comment}"
        i += 1
    return [s.replace("nan", "") for s in new_annotations[:8]] + [secondary_structure] + [s.replace("nan", "") for s in new_annotations[8:]]


def rollup(search_engine: str, search_engine_filepath: str, use_target_list: bool, 
           target_list_filepath: str, max_missed_cleavages: int, protease: str, 
           fdr_threshold: float, use_quant: bool, user_PTMs: list, 
           proteome_fasta_filepath: str, intensity_method: str, add_annotation: bool, 
           species_id: str, output_filename: str) -> int:
    """starts proteoSushi rollup when called by run_proteoSushi

    Arguments:
        search_engine {str} -- "maxquant", "mascot", or "generic"
        search_engine_filepath {str} -- the filepath for search engine output
        use_target_list {bool} -- whether to use the target list to prioritize matches
        target_list_filepath {str} -- the filepath for the target gene list
        max_missed_cleavages {int} -- the maximum allowed missed cleavages in a peptide
        protease {str} -- the protease used previously to cleave the proteins in the sample
        fdr_threshold {float} -- threshold used for the pep_expect and PEP columns in mascot/maxquant
        use_quant {bool} -- whether to sum/average the quant values
        user_PTMs {list} -- a list of PTMs that will be used in analysis
        proteome_fasta_filepath {str} -- the filepath of the proteome fasta file
        intensity_method {str} -- whether to "sum" or "average" the peaks for combined peptides
        add_annotation {bool} -- whether to query uniprot and add the annotation onto the results
        species_id {str} -- the species ID (e.g. 9606) to get the annotation score dictionary from Uniprot
    Returns:
        int -- possible error flag
    """
    threshold = fdr_threshold
    use_target = use_target_list
    use_intensities = use_quant
    if species_id != "":
        annot_filename = download_AS_file(species_id)
    else:
        annot_filename = ""
    rollup_file = search_engine

    if rollup_file == "generic":
        #output_filename = "generic_rollup.csv"
        sequence_index, modified_sequence_index, mod_dict, intensity_start, data_filename, \
            var_mod_dict = compile_data_generic(search_engine_filepath, user_PTMs)
        data_file = open(data_filename, 'r')
        tsv_reader = csv.reader(data_file, quotechar='"')
    elif rollup_file == "maxquant":
        #output_filename = "maxquant_rollup.csv"
        sequence_index, modified_sequence_index, mod_dict, intensity_start, data_filename, \
            var_mod_dict = compile_data_maxquant(search_engine_filepath, user_PTMs)
        data_file = open(data_filename, 'r')
        tsv_reader = csv.reader(data_file, delimiter='\t', quotechar='"')
    elif rollup_file == "mascot":
        #output_filename = "mascot_rollup.csv"
        sequence_index, modified_sequence_index, mod_dict, intensity_start, data_filename, \
            var_mod_dict = compile_data_mascot(search_engine_filepath, user_PTMs)
        data_file = open(data_filename, 'r')
        tsv_reader = csv.reader(data_file, quotechar='"')
        if intensity_start is None and use_quant:
            return 2
    else:
        assert False, "Not a valid file from search"

    enzyme = protease
    missed_cleavages = max_missed_cleavages
    pep_dict = load_pepdict(proteome_fasta_filepath, enzyme, missed_cleavages)
    annotDict = __load_annot_dict(annot_filename)

    if use_target:
        target_genes = open(target_list_filepath, 'r').readlines()
    else:
        target_genes = []

    header = next(tsv_reader)
    # Use the PEP column if the user chose Maxquant output
    if rollup_file == "maxquant":
        false_disc = header.index("PEP")
    intensity_dict = dict()
    gene_results = list()
    sparql_input = list()
    unmatched_peps = 0
    missing_PTM = 0
    total_seqs = 0
    over_threshold = 0
    unmatched_sequences = []

    for row in tsv_reader:
        # If a maxquant file is used and the false disc rate of this peptide is not below the threshold, skip it
        if rollup_file == "maxquant" and not threshold is None and float(row[false_disc]) > threshold:
            over_threshold += 1
            continue
        total_seqs += 1
        raw_seq = row[sequence_index]
        #if raw_seq == "DLGGIVLANACGPCIGQWDR":
        #    print("start")
        pep_mod_seq = row[modified_sequence_index]
        #if pep_mod_seq == "DLGGIVLANAC(ne)GPC(ca)IGQWDR":
        #    print("start")
        pep_seq = row[sequence_index].replace("L","I")
        if pep_seq is None:
            print("HALT!")  # A handled exception would be preferable here
            break
        genes_positions = pep_dict.get(pep_seq)
        if use_intensities and rollup_file == "generic":
            intensities = [e for i, e in enumerate(row) if i in intensity_start]
            # If there is no intensity, skip that site
            if intensities == '' or intensities[0] == '':
                continue
            intensity_header = [e for i, e in enumerate(header) if i in intensity_start]
        elif rollup_file == "maxquant":
            pep_mod_seq = pep_mod_seq.strip('_')

            if use_intensities:
                #intensities = row[intensity_start]
                intensities = [e for i, e in enumerate(row) if i in intensity_start]
                # If there is no intensity, skip that site
                if intensities == '' or intensities[0] == '':
                    continue
                #intensities = [intensities]
                intensity_header = [e for i, e in enumerate(header) if i in intensity_start]
        elif rollup_file == "mascot":
            if pep_mod_seq == "":
                continue
            new_seq = pep_seq
            pep_mod_seq = pep_mod_seq.split('.')[1]
            i = len(pep_mod_seq) - 1
            inv_mod_dict = {v:k for k, v in var_mod_dict.items()}
            while i >= 0:
                if pep_mod_seq[i] != "0":  # If there is a mod
                    new_seq = new_seq[:i+1] + '(' + inv_mod_dict[pep_mod_seq[i]] + ')' + new_seq[i+1:]
                i -= 1
            pep_mod_seq = new_seq

            if use_intensities:
                if len(row) < len(header):  # If the row is cut short (of the intensity cells), skip to the next one
                    continue
                while intensity_start < len(header) and '/' in row[intensity_start]:
                    intensity_start += 2  # 
                # NOTE: This is now an error to the GUI and this line of code shouldn't run.
                assert intensity_start < len(header), "Mascot intensity values may be missing, please check"

                intensities = row[intensity_start+1::2]
                # If there is no intensity, skip that site
                if intensities[0] == "---":  # If there are no intensity values in this line, go to the next one
                    continue
                intensity_header = row[intensity_start::2]  # Ideally, this would happen outside of this loop

        new_user_PTMs = user_PTMs
        if rollup_file == "maxquant":
            new_user_PTMs = [ptm.lower()[:2] for ptm in user_PTMs]
        new_pep_mod_seq = __clean_pep_seq(cleave_rules[enzyme], pep_mod_seq, new_user_PTMs)
        if genes_positions and len(genes_positions) == 1:
            gene, start_pos, unpid, protein_name = list(genes_positions)[0]
            if not new_pep_mod_seq in mod_dict:
                print("\033[91m {}\033[00m".format(f"{pep_seq} does not have the PTM(s) selected!"))
                missing_PTM += 1
                continue
            mods = mod_dict[new_pep_mod_seq]
            if not (any(mods) and set([m[0] for m in mods]) & set(user_PTMs)): 
                continue
            for mod in list(set(mods)):
                if not mod[0] in user_PTMs:
                    continue
                site = start_pos + mod[1] + 1  # The last +1 is to change from 0-indexing to 1-indexing (like humans use)
                #if gene.upper() == "ACTL6A":
                #    print("ACTL6A")
                if use_intensities:  # If the user chose to combine/average intensities
                    intensity_dict, to_add = __add_intensity(intensity_dict, 
                                                             new_pep_mod_seq, 
                                                             [gene], site, intensities)
                else:
                    key = f"{new_pep_mod_seq}|{gene.upper()}|{str(site)}"
                    if key in intensity_dict:
                        to_add = False
                    else:
                        to_add = True
                        intensity_dict[key] = 0
                if to_add:
                    if use_target and gene.upper() + '\n' in target_genes:  # TODO: fetch the uniprot ID from the match
                        gene_results.append([
                            gene, 
                            site, 
                            protein_name,
                            "", 
                            gene, 
                            raw_seq, 
                            new_pep_mod_seq, 
                            annotDict[unpid] if unpid in annotDict else "", 
                            unpid
                            ])
                        if len(unpid) >= 5:
                            sparql_input.append(tuple((unpid, site, gene)))
                    else:
                        gene_results.append([
                            gene, 
                            site, 
                            protein_name,
                            "", 
                            "", 
                            raw_seq, 
                            new_pep_mod_seq, 
                            annotDict[unpid] if unpid in annotDict else "", 
                            unpid
                            ])  # NOTE: I changed this from a tuple, so things might be different
                        if len(unpid) >= 5:
                            sparql_input.append(tuple((unpid, site, gene)))
        elif genes_positions and len(genes_positions) > 1:
            isTarget, match = __chooseHit(genes_positions, target_genes, annotDict, use_target)
            # If there was > 1 target genes, non-target genes, or a combination, AND none was chosen
            if match is None:  
                continue
            if not match[0] is None:  # Checks to see that there has been a match
                if isinstance(match, tuple):  # Checks if this is just a single match.
                    gene = match[0]
                    if not new_pep_mod_seq in mod_dict:
                        print("\033[91m {}\033[00m".format(f"{pep_seq} does not have the PTM(s) selected!"))
                        missing_PTM += 1
                        continue
                    mods = mod_dict[new_pep_mod_seq]
                    for mod in list(set(mods)):
                        if not mod[0] in user_PTMs:
                            continue
                        site = match[1] + mod[1] + 1  # The last +1 is to change from 0-indexing to 1-indexing (like humans use)
                        #if gene.upper() == "ACTL6A":
                        #    print("ACTL6A")
                        to_add = None
                        if use_intensities:
                            intensity_dict, to_add = __add_intensity(intensity_dict, 
                                                                     new_pep_mod_seq, 
                                                                     [match[0]], 
                                                                     site, intensities)
                        else:
                            key = f"{new_pep_mod_seq}|{gene.upper()}|{str(site)}"
                            if key in intensity_dict:
                                to_add = False
                            else:
                                to_add = True
                                intensity_dict[key] = 0
                        if to_add:
                            if isTarget:
                                gene_results.append([
                                    gene, 
                                    site, 
                                    match[3],
                                    "", 
                                    gene, 
                                    raw_seq, 
                                    new_pep_mod_seq, 
                                    annotDict[match[2]] if match[2] in annotDict else "", 
                                    match[2]
                                    ])
                                if len(match[2]) >= 5:
                                    sparql_input.append(tuple((match[2], site, gene)))
                            else:
                                gene_results.append([
                                    gene, 
                                    site, 
                                    match[3],
                                    "", 
                                    "", 
                                    raw_seq, 
                                    new_pep_mod_seq, 
                                    annotDict[match[2]] if match[2] in annotDict else "", 
                                    match[2]
                                    ])
                                if len(match[2]) >= 5:
                                    sparql_input.append(tuple((match[2], site, gene)))
                else:  # There are multiple matches
                    if not new_pep_mod_seq in mod_dict:
                        print("\033[91m {}\033[00m".format(f"{pep_seq} does not have the PTM(s) selected!"))
                        missing_PTM += 1
                        continue
                    mods = mod_dict[new_pep_mod_seq]
                    for mod in list(set(mods)):
                        if not mod[0] in user_PTMs:
                            continue
                        additGenes = list()
                        for tup in match:
                            additGenes.append(tup[0])
                        assert len(additGenes) > 1, "The # of matches should be >1, but isn't"
                        additGenes = list(set(additGenes))
                        #gene_true_pos = f"{match[0][0]}|{match[0][1]+mod[1]}"
                        site = match[0][1] + mod[1] + 1  # The last +1 is to change from 0-indexing to 1-indexing (like humans use)
                        if use_intensities:
                            intensity_dict, to_add = __add_intensity(intensity_dict,
                                                                     new_pep_mod_seq, 
                                                                     additGenes, 
                                                                     site, intensities)
                        else:
                            if any(f"{new_pep_mod_seq}|{gene.upper()}|{str(site)}" in intensity_dict for gene in additGenes):
                                to_add = False
                            else:
                                to_add = True
                                for gene in additGenes:
                                    intensity_dict[f"{new_pep_mod_seq}|{gene.upper()}|{str(site)}"] = 0
                        if to_add:
                            for addit_gene in additGenes:
                                #additGenes.remove(match[0][0])
                                gene_list = [x for x in additGenes if x != addit_gene]
                                current_match = [x for x in match if x[0] == addit_gene][0]
                                if isTarget:
                                    gene_results.append([
                                        addit_gene, 
                                        site,
                                        current_match[3], 
                                        #match[0][3],
                                        ' '.join(gene_list), 
                                        ' '.join([i[0] for i in match]), 
                                        raw_seq, 
                                        new_pep_mod_seq, 
                                        annotDict[current_match[2]] if current_match[2] in annotDict else "", 
                                        ' '.join([i[2] for i in match])
                                        ])
                                    if len(current_match[2]) >= 5:
                                        sparql_input.append(tuple((current_match[2], site, current_match[0])))
                                else:
                                    gene_results.append([
                                        addit_gene, 
                                        site, 
                                        current_match[3], 
                                        ' '.join(gene_list), 
                                        "", 
                                        raw_seq, 
                                        new_pep_mod_seq, 
                                        annotDict[current_match[2]] if current_match[2] in annotDict else "", 
                                        ' '.join([i[2] for i in match])
                                        ])
                                    if len(current_match[2]) >= 5:
                                        sparql_input.append(tuple((current_match[2], site, current_match[0])))
            else:  # NOTE: it generally shouldn't get here
                unmatched_peps += 1
                unmatched_sequences.append(tuple([raw_seq, pep_mod_seq]))
        else:  # There was no match for the peptide in the pepdict
            unmatched_peps += 1
            unmatched_sequences.append(tuple([raw_seq, pep_mod_seq]))
    data_file.close()
    # Prints the stats from the rollup
    print("\033[93m {}\033[00m".format(f"\nUnmatched Peptides: {unmatched_peps}\nMissing Selected PTMs: {missing_PTM}\nTotal Peptides: {total_seqs}"))
    '''
    ########################
    #Just to annotate EGFR #
    ########################
    if add_annotation and False:
        sparql_dict = dict()
        #with open("EGFR_annotation_uniprot_unfixed.csv", 'r') as annotation_file:
        
        with open("sparql-complete.csv", 'r') as annotation_file:
            annotations_full = pd.DataFrame(columns=annotation_file.readline().split(','))
            #print("here")
            for line in annotation_file:
                if line[:5] == "entry":
                    continue
                else:
                    for line_split in csv.reader([line], delimiter=',', quotechar='"'):
                        annotations_full = annotations_full.append(pd.Series(line_split, index=annotations_full.columns), ignore_index=True)
        
        annotations_full = pd.read_csv("sparql-complete.csv")
        #annotations_full = pd.read_json("sparql-complete.srj")
        annotations_full = annotations_full[annotations_full.entry != "entry"]
        sparql_output, sparql_dict = sparql.process_sparql_output(annotations_full, sparql_dict)
    '''
    # If the user chose, it combines the annotation onto the rollup results (eventually)
    if add_annotation:
        print("\033[95m {}\033[00m".format("\nQuerying Uniprot for Annotations!"))
        batch = 50
        i = 0
        results_annotated = 0
        sparql_output_list = list()
        sparql_input = sorted(list(set(sparql_input)), key=lambda x: x[2])
        sparql_dict = dict()
        # Separates the input into batches then sends those batches
        while i + batch <= len(sparql_input):
            # Makes the request and sends it to uniprot
            batch_output = sparql_request(sparql_input[i:i+batch])

            #If there is a 502 error, send that to the GUI to display
            if isinstance(batch_output, int) and batch_output == 502:
                return 502

            # If after all attempts to get annotations for this batch has failed, this is reported and the next batch will be sent
            if batch_output is None or (not isinstance(batch_output, str) and batch_output.empty):
                print("\033[91m {}\033[00m".format(f"Lines {i+2} to {i+batch+1} not annotated!"))
                i += batch
                continue

            # This processes and combines the annotations to 1 per site
            sparql_output, sparql_dict = process_sparql_output(batch_output, sparql_dict)
            if not sparql_output:
                print("\033[91m {}\033[00m".format(f"Lines {i+2} to {i+batch+1} not annotated!"))
                i += batch
                continue
            sparql_output_list.append(sparql_output)
            i += batch
            results_annotated += batch
            print("\033[96m {}\033[00m" .format(f"{round(float(results_annotated)/len(sparql_input)*100, 2)}% of results annotated"))
        batch_output = sparql_request(sparql_input[i:])

        #If there is a 502 error, send that to the GUI to display
        if isinstance(batch_output, int) and batch_output == 502:
            return 502

        if batch_output is None or (not isinstance(batch_output, str) and batch_output.empty):
            print("\033[91m {}\033[00m".format(f"Lines {i+2} to {i+len(sparql_input[i:])+1} not annotated!"))
            pass
        else:
            sparql_output, sparql_dict = process_sparql_output(batch_output, sparql_dict)
            if not sparql_output:
                print("\033[91m {}\033[00m".format(f"Lines {i+2} to {i+batch+1} not annotated!"))
                pass
            else:
                sparql_output_list.append(sparql_output)
                results_annotated += len(sparql_input[i:])
        print("\033[92m {}\033[00m" .format(f"\n{round(float(results_annotated)/len(sparql_input)*100, 2)}% of results annotated"))
        # Writes the annotations to a separate file in case the user wants to view those in a more vertical way
        with open("sparql_annotations.csv", 'w') as spql_annot:
            out_writer = csv.writer(spql_annot)
            for line in sparql_output_list:
                out_writer.writerow(line)
    print("\033[95m {}\033[00m".format("\nWriting the rollup output file"))

    # Prints out the completed rollup with annotations from Uniprot (if requested)
    with open(output_filename, 'w', newline = '') as w1:
        out_writer = csv.writer(w1)
        header2 = ["Gene", "Site", "Protein_Name", "Shared_Genes", "Target_Genes", "Peptide_Sequence", 
            "Peptide_Modified_Sequence", "Annotation_Score", "Uniprot_Accession_ID"]
        if use_intensities:
            header2 += [ih + f" ({intensity_method})" for ih in intensity_header]
        if add_annotation:
            # TODO: Change the header to be just one group of annotations
            '''
            annotations_length = max(len(v) for k, v in sparql_dict.items())
            if annotations_length > 0:
                header2 += ["entry", "position", "lengthOfSequence"]
                annotations_length -= 3
            while annotations_length > 0:
                header2 += ["begin", "end", "regionOfInterest", "catalyicActivity", "location", 
                            "ec", "rhea", "type", "comment"]
                annotations_length -= 9
            '''
            header2 += ["Length_Of_Sequence", "Range_of_Interest", "Region_of_Interest", 
                        "Subcellular_Location", "Enzyme_Class", "rhea", "Secondary_Structure", 
                        "Active_Site_Annotation", "Alternative_Sequence_Annotation", 
                        "Chain_Annotation", "Compositional_Bias_Annotation", 
                        "Disulfide_Bond_Annotation", "Domain_Extent_Annotation", 
                        "Lipidation_Annotation", "Metal_Binding_Annotation", 
                        "Modified_Residue_Annotation", "Motif_Annotation", 
                        "Mutagenesis_Annotation", "Natural_Variant_Annotation", 
                        "NP_Binding_Annotation", "Other", "Region_Annotation", "Repeat_Annotation", 
                        "Topological_Domain_Annotation", "Zinc_Finger_Annotation"]
        out_writer.writerow(header2)
        # This builds the rollup output file depending on what the user chose
        for i in sorted(gene_results, key=lambda r: r[0]):
            if rollup_file == "maxquant":
                if not any(ptm[:2].lower() in i[6] for ptm in user_PTMs):
                    continue
            elif not any(ptm in i[6] for ptm in user_PTMs):  # If none of the chosen ptms are in the rollup line
                continue
            # Start by adding in the base data
            writable_row = i
            if use_intensities:  # Add to that the intensity data if requested
                if intensity_method == "sum":  # Reports the sum of each peak
                    writable_row += intensity_dict[f"{i[6]}|{i[0].upper()}|{i[1]}"][:-1]
                elif intensity_method == "average":  # Calculates the average for each peak and reports
                    N = intensity_dict[f"{i[6]}|{i[0].upper()}|{i[1]}"][-1]
                    intensities = intensity_dict[f"{i[6]}|{i[0].upper()}|{i[1]}"][:-1]
                    writable_row += [float(x)/N for x in intensities]
            if add_annotation:
                try:
                    compressed_annotations = __compress_annotations(sparql_dict[i[8].split(' ')[0] + '|' + str(i[1])])
                    writable_row += compressed_annotations[2:]#5]# + compressed_annotations[6:]
                except KeyError:
                    #print(i[6] + '|' + str(i[1]) + " not in dict")
                    pass
            out_writer.writerow(writable_row)
        #print(sparql_dict)

    # Puts all of the unmatched sequences into a new file
    with open(f"unmatched_sequences_{rollup_file}.csv", 'w', newline = '') as w2:
        out_writer = csv.writer(w2)
        header = ["Sequence", "Peptide Modified Sequence"]
        out_writer.writerow(header)
        for i in unmatched_sequences:
            writable_row = list(i)
            out_writer.writerow(writable_row)
#EOF


"""Quick file to make sure merge is working as I expect"""
import pandas as pd
df_region = pd.read_csv("test_region_dataframe.csv")
df_subcell = pd.read_csv("test_subcell_dataframe.csv")
df_catalytic = pd.read_csv("test_catalytic_dataframe.csv")
df_extras = pd.read_csv("test_extras_dataframe.csv")

df = df_region
df = df.merge(df_subcell, how="outer", on="entry")
df.to_csv("test_merge_1.csv")
df = df.merge(df_catalytic, how="outer", on=["entry", " position"])
df.to_csv("test_merge_2.csv")
df = df.merge(df_extras, how="outer", on=["entry", " position"])
df.to_csv("test_merge_3.csv")
#EOF


"""parse_proteome.py: parses the proteome FASTA file to make sure it is in an acceptable format
and grabs the species ID"""

import os

def fasta_entry_check(header: str, sequence: str) -> str:
    """Checks whether the header has organism info and returns as tuple

    Arguments:
        header {str} -- header line
        sequence {str} -- protein sequence
    Returns:
        str -- species ID
    """
    unpid = "N/A"
    # Check organism
    if 'OX=' in header:
        organism = int(header.split('OX=')[1].split()[0])
    elif "OS=" in header:
        organism = header.split('OS=')[1].split(" GN=")[0]
    else:
        return None  # skip altogether if no organism info in line
    # Attempts to grab the Uniprot ID from the header
    if header.count('|') >= 2:
        unpid = header.split('|')[1]

    if unpid == "N/A":
        return "ERROR"

    return str(organism)

def parse_proteome(filename: str) -> list:
    """checks FASTA file to make sure it is in the correct format and grabs the species ID

    Arguments:
        filename {str} -- the filepath+name of the FASTA file
    Returns:
        bool -- whether an error occurred during parsing (e.g. incorrect format)
        str -- the species ID from the fasta file
    """
    species_IDs = list()
    if os.path.exists(filename):
        with open(filename) as input_file:
            print(f"Checking proteome file: {filename}")
            header = None
            seq = ''
            for line in input_file:
                if line.startswith('>') and seq:
                    species_ID = fasta_entry_check(header, seq)
                    if species_ID == "ERROR":
                        return True, None
                    elif species_ID is None:
                        pass
                    else:
                        species_IDs.append(species_ID)
                    header = line.strip()
                    seq = ''
                elif line.startswith('>'):  # first entry of file
                    header = line.strip()
                else:
                    seq += line.strip()
            fasta_entry_check(header, seq)  # For last entry
            if species_ID == "ERROR":
                return True, None
            elif species_ID is None:
                pass
            else:
                species_IDs.append(species_ID)

    # If there are no species listed, it will send back None
    if not species_IDs:
        return False, None
    # Grabs the most common of the species IDs and uses that
    return False, max(set(species_IDs), key=species_IDs.count)
#EOF